require 'spec_helper'

module Hikki
  module Adapters
    describe RedisAdapter do
      subject(:adapter) { RedisAdapter.new(connection, uuid_generator) }
      let(:connection) { double :connection, sadd: true, smembers: [], hset: true, hget: true, hdel: true, del: true }
      let(:uuid_generator) { double :uuid_generator, uuid: '12345' }
      let(:collection) { 'collection1' }

      describe '#index' do
        it 'returns true' do
          expect(adapter.index(collection, :field1)).to be_true
        end

        it 'adds the field to the index list' do
          adapter.index(collection, :field1)
          expect(connection).to have_received(:sadd).with('hikki:collection1:indexes', 'field1')
        end
      end

      describe '#save' do
        context 'when an id is provided in the data' do
          let(:data) { { id: id, field1: 'test', field2: 123 } }
          let(:expected) { { 'id' => id, 'field1' => 'test', 'field2' => 123 } }
          let(:json) { '{"id":"1","field1":"test","field2":123}' }
          let(:id) { '1' }

          it 'returns the data' do
            expect(adapter.save(collection, data)).to eq expected
          end

          it 'persists the data in the store' do
            adapter.save(collection, data)
            expect(connection).to have_received(:hset).with('hikki:collection1', id, json)
          end
        end

        context 'when an id is not provided in the data' do
          let(:data) { { field1: 'test', field2: 123 } }
          let(:expected) { { 'id' => id, 'field1' => 'test', 'field2' => 123 } }
          let(:json) { '{"field1":"test","field2":123,"id":"12345"}' }
          let(:id) { '12345' }

          it 'returns the data with the id added' do
            expect(adapter.save(collection, data)).to eq expected
          end

          it 'persists the data in the store' do
            adapter.save(collection, data)
            expect(connection).to have_received(:hset).with('hikki:collection1', id, json)
          end
        end

        context 'when an index exists on a field in the data' do
          let(:data) { { id: id, field1: 'test', field2: 123 } }
          let(:json) { '{"id":"1","field1":"test","field2":123}' }
          let(:id) { '1' }
          before do
            connection.stub(:smembers).with('hikki:collection1:indexes').and_return(['field1'])
            connection.stub(:hget).with('hikki:collection1:field1', 'test').and_return('[]')
          end

          it 'adds an entry in the index' do
            expect(connection).to receive(:hset).with('hikki:collection1', id, json)
            expect(connection).to receive(:hset).with('hikki:collection1:field1', 'test', '["1"]')
            adapter.save(collection, data)
          end
        end
      end

      describe '#find' do
          let(:id) { '1' }

        context 'when the id exists' do
          let(:data) { { id: id, field1: 'test', field2: 123 } }
          let(:json) { '{"id":"1","field1":"test","field2":123}' }
          let(:expected) { { 'id' => id, 'field1' => 'test', 'field2' => 123 } }
          before { connection.stub(:hget).with('hikki:collection1', id).and_return(json) }

          it 'retrieves the data' do
            expect(adapter.find(collection, id)).to eq expected
          end
        end

        context 'when the id does not exist' do
          let(:expected) { {} }
          before { connection.stub(:hget).with('hikki:collection1', id).and_return(nil) }

          it 'returns an empty hash' do
            expect(adapter.find(collection, id)).to eq expected
          end
        end
      end

      describe '#all' do
        context 'with a record in the collection' do
          let(:data) { { id: '1', field1: 'test', field2: 123 } }
          let(:json) { '{"id":"1","field1":"test","field2":123}' }
          let(:expected) { { 'id' => '1', 'field1' => 'test', 'field2' => 123 } }
          before { connection.stub(:hvals).with('hikki:collection1').and_return([json]) }

          it 'returns an array containing the record' do
            expect(adapter.all(collection)).to eq [expected]
          end
        end

        context 'with no records in the collection' do
          before { connection.stub(:hvals).with('hikki:collection1').and_return([]) }

          it 'returns an array containing the record' do
            expect(adapter.all(collection)).to eq []
          end
        end

        context 'with multiple records and paging options specified' do
          before do
            results = (1..10).map {|i| "{\"id\":\"#{i}\",\"a\":1}" }
            connection.stub(:hvals).with('hikki:collection1').and_return(results)
          end

          context 'with limit 2' do
            let(:limit) { 2 }

            it 'returns only 2 records' do
              expect(adapter.all(collection, {limit: 2}).count).to eq 2
            end

            context 'with an offset of 4' do
              it 'returns the 2 records starting with the offset' do
                expect(adapter.all(collection, {limit: 2, offset: 4})).to eq [{'id' => '5', 'a' => 1}, {'id' => '6', 'a' => 1}]
              end
            end
          end
        end
      end

      describe '#find_by' do
        context 'with a record in the collection matching' do
          let(:data) { { id: id, field1: 'test', field2: 123 } }
          let(:json) { '{"id":"1","field1":"test","field2":123}' }
          let(:expected) { { 'id' => id, 'field1' => 'test', 'field2' => 123 } }
          let(:id) { '1' }

          context 'without an index' do
            before do
              connection.stub(:sismember).with('hikki:collection1:indexes', 'field2').and_return(false)
              connection.stub(:hvals).with('hikki:collection1').and_return([json])
            end

            it 'returns an array containing the record' do
              expect(adapter.find_by(collection, :field2, 123)).to eq [expected]
            end
          end

          context 'with an index' do
            before do
              connection.stub(:sismember).with('hikki:collection1:indexes', 'field1').and_return(true)
              connection.stub(:hget).with('hikki:collection1:field1', 'test').and_return('[1]')
              connection.stub(:hget).with('hikki:collection1', id).and_return(json)
              connection.stub(:pipelined).and_yield.and_return([json])
            end

            it 'returns an array containing the record' do
              expect(adapter.find_by(collection, :field1, 'test')).to eq [expected]
            end
          end
        end

        context 'with no matching records in the collection' do
          before do
            connection.stub(:sismember).with('hikki:collection1:indexes', 'field1').and_return(false)
            connection.stub(:hvals).with('hikki:collection1').and_return([])
          end

          it 'returns an empty array' do
            expect(adapter.find_by(collection, :field1, 'foo')).to eq []
          end
        end

        context 'with multiple matching records and paging options specified' do
          before do
            results = (1..10).map {|i| "{\"id\":\"#{i}\",\"a\":1}" }
            connection.stub(:sismember).with('hikki:collection1:indexes', 'a').and_return(false)
            connection.stub(:hvals).with('hikki:collection1').and_return(results)
          end

          context 'with limit 2' do
            let(:limit) { 2 }

            it 'returns only 2 records' do
              expect(adapter.find_by(collection, :a, 1, {limit: 2}).count).to eq 2
            end

            context 'with an offset of 4' do
              it 'returns the 2 records starting with the offset' do
                expect(adapter.find_by(collection, :a, 1, {limit: 2, offset: 4})).to eq [{'id' => '5', 'a' => 1}, {'id' => '6', 'a' => 1}]
              end
            end
          end
        end
      end

      describe '#remove' do
        context 'with a record for the id in the collection' do
          let(:id) { '1' }
          let(:json) { '{"id":"1","field1":"test","field2":123}' }
          before { connection.stub(:hget).with('hikki:collection1', id).and_return(json) }

          it 'returns true' do
            expect(adapter.remove(collection, id)).to be_true
          end

          it 'removes the record from the store' do
            adapter.remove(collection, id)
            expect(connection).to have_received(:hdel).with('hikki:collection1', id)
          end

          context 'when there is an index' do
            before do
              connection.stub(:smembers).with('hikki:collection1:indexes').and_return(['field1'])
              connection.stub(:hget).with('hikki:collection1:field1', 'test').and_return('["1"]')
            end

            it 'removes the id from the index' do
              adapter.remove(collection, id)
              expect(connection).to have_received(:hset).with('hikki:collection1:field1', 'test', '[]')
            end
          end
        end
      end

      describe '#remove_all' do
        context 'with a record in the collection' do
          it 'returns true' do
            expect(adapter.remove_all(collection)).to be_true
          end

          it 'removes all records from the store' do
            expect(connection).to receive(:del).with('hikki:collection1:indexes')
            expect(connection).to receive(:del).with('hikki:collection1')
            adapter.remove_all(collection)
          end

          context 'when there is an index' do
            before { connection.stub(:smembers).with('hikki:collection1:indexes').and_return(['field1']) }

            it 'removes the indexes' do
              adapter.remove_all(collection)
              expect(connection).to have_received(:del).with('hikki:collection1:field1')
              expect(connection).to have_received(:del).with('hikki:collection1:indexes')
            end
          end
        end
      end
    end
  end
end
